package vishnu.builder;

import java.io.*;
import java.util.*;
import java.util.Hashtable;
import vishnu.util.*;

public class CKIndexer extends Indexer
{

    static private String delimiter = "()<>.,?!'\" "; 


    private static void usage()
    {
        System.out.println("java Indexer/CKIndexer \n -i InputDirectory [$COLL_HOME/import/]) \n -o OutputDirectory [$COLL_HOME/index/ck_index/java] \n -r root directory of collection \n -list (name of file listing the docs to be indexed)[docs.lst]\n -d string containing chars to be eliminated from words\n");
	System.exit(0);
    }
    
    
    public static void main(String[] args)
    {
        String in = null;
        String out = null;
        //String COLL_HOME = null;
	//String docList = null;
	
        if( args.length == 0 ){
            usage();
            System.exit(0);
        }
	
        int a = 0;
        while( a < args.length ){
            if( args[a].equals("-i") )
                in = args[++a];
            else if( args[a].equals("-o") )
                out = args[++a];
            //else if( args[a].equals("-r") )
            //    COLL_HOME = args[++a];
	    //else if( args[a].equals("-list") )
	    //	docList = args[++a];
	    else if( args[a].equals("-d") )
		delimiter = args[++a];
            a++;
        }
	
	CKIndexer cki = new CKIndexer();

	if (out==null || in==null) {
	    usage();
	    System.exit(0);
	}
	
	//if( COLL_HOME == null ) usage();
	//cki.setRootDirectory(COLL_HOME);
	
	//if( out == null ) out = COLL_HOME + "/index/ck_index/java/";
	cki.setOutputDirectory(out);
	
	//if( in == null ) in = COLL_HOME + "/import/";
	cki.setInputDirectory(in);
	
	//if( docList != null ) cki.setDocList(docList);
	
	cki.startIndexing();
    }
    
    
    public CKIndexer(){
	super("CKIndexer");
    }
    
    
//     public void setDocList(String doclist)
//     {
// 	docList = doclist;
//     }


    public void startIndexing(){
	
	initialize();
	/**** read in archive ****/
	
	Vector docs = getDocList(inputDir); //readDocList(docList);
	
	System.out.println(docs.size() + " documents read");
	
	/****  hash of ALL words of corpus + their document frequency ****/
	
	Hashtable corpus_hash = new Hashtable();
	
	
	/**** readDocuments and populate hash ****/

	System.out.println("First run: getting document frequencies of all words");
	 
	getDocFreqs(docs, corpus_hash);
	
	
	/**** prune words to get candidate keywords ****/

	int min = 3;
	int max = (int)(0.3*docs.size());
	int minLength = 4;
	
	int num_ck = pruneKeywords(corpus_hash, minLength, min, max);
	
	
	/**** sort hashtable by value ****/
	
	Enumeration enum = corpus_hash.keys();
	String[] words_ck = new String[num_ck];
	int[] freqs_ck = new int[num_ck];
	
	int count = 0;
	while( enum.hasMoreElements() ){
	    words_ck[count] = (String)enum.nextElement();
	    freqs_ck[count] = -((Integer)corpus_hash.get(words_ck[count])).intValue();
	    count++;
	}
	
	Misc.quickSort(freqs_ck,words_ck);
	
	
	/**** normalize document frequencies ****/
	
	float[] freqs_f = new float[num_ck];
	
	for( int i = 0; i < num_ck; i++ ){
	    freqs_f[i] = (float)(-1*freqs_ck[i])/docs.size();
	}
	
	
	/**** save document frequencies and candidate keywords ****/
	
	printToFile(freqs_f,outputDir + "doc_freq");
	printToFile(words_ck,outputDir + "ck_words");	
	
	
	/**** create (ck_word, index hash) ****/
	
	Hashtable ck_hash = new Hashtable();
	
	for( int i = 0; i < words_ck.length; i++ ){
	    ck_hash.put(words_ck[i],new Integer(i));
	}
	
	
	/**** read documents and store for each document the ck words it contains and their frequencies ****/

	System.out.println("Second run: getting term frequencies of candidate keywords");
	
	Vector termFreq = getTermFreqs(docs,ck_hash);
	
	
	/**** save term frequency vector ****/ 
	
	printToFile(termFreq,outputDir + "term_freq");
    }
    
    
    private int pruneKeywords(Hashtable corpus_hash, int minLength, int min, int max){
	Set ck_set = corpus_hash.keySet();
	Enumeration enum = corpus_hash.keys();
	
	int count = 0;
	int pcount = 0;
      	
	while( enum.hasMoreElements() ) {
	    
	    String word = (String)enum.nextElement();
	    int freq = ((Integer)corpus_hash.get(word)).intValue();
	    
	    
	    /**** if word is a number remove ****/
	    
	    try{
		int i = Integer.parseInt(word);
		corpus_hash.remove(word);
		ck_set.remove(word);
		continue;
	    }catch(NumberFormatException ex){}


	    /**** if word too rare or too freq remove ****/
	    
	    if( word.length() < minLength || freq < min || freq > max ){
		corpus_hash.remove(word);
		ck_set.remove(word);
		continue;
	    }
	}
	return corpus_hash.size();
    }
    
    
    private Vector getTermFreqs(Vector docs, Hashtable ck_hash)
    {
	Vector termFreqVector = new Vector();
	
	for( int i = 0; i < docs.size(); i++ ){

	    Hashtable doc_ck_hash = new Hashtable();
	    String doc = (String)docs.elementAt(i);
	    
	    try{
		BufferedReader dis = new BufferedReader(new InputStreamReader(new FileInputStream(doc)));
		String line = "";
		StringTokenizer strtok;
		
		while( (line=dis.readLine())!=null ){
		    
		    /**** read word by word ****/
		    
		    strtok = new StringTokenizer(line,delimiter);
		    
		    while( strtok.hasMoreTokens() ){
			String word = strtok.nextToken();
			Integer index = null;
			if( (index = (Integer)ck_hash.get(word)) != null ){
			    Integer count;
			    
			    if( (count = (Integer)doc_ck_hash.get(index)) != null ){
				int newCount = count.intValue() + 1;
				doc_ck_hash.put(index,new Integer(newCount));
			    }else{
				doc_ck_hash.put(index,new Integer(1));				
			    }
			}
		    }
		}
		dis.close();
	    }catch(Exception ex){ex.printStackTrace();}
	
	
	    
	    termFreqVector.addElement(doc_ck_hash);
	}
	
	return termFreqVector;
    }
    

    private void getDocFreqs(Vector docs, Hashtable corpus_hash)
    {
	for( int i = 0; i < docs.size(); i++ ){
	    
	    String doc = (String)docs.elementAt(i);
	    
	    Hashtable doc_hash = new Hashtable();
	    
	    try{
		BufferedReader dis = new BufferedReader(new InputStreamReader(new FileInputStream(doc)));
		String line = "";
		StringTokenizer strtok;
		
		while( (line=dis.readLine())!=null ){
		    
		    
		    /**** read word by word ****/
		    
		    strtok = new StringTokenizer(line,delimiter);
		    
		    while( strtok.hasMoreTokens() ){
			String word = strtok.nextToken();
			
			
			if( (Object)doc_hash.get(word) == null ){
			    
			    
			    /**** make a note that there is at least one occurrence ****/
			    
			    doc_hash.put(word,word);
			    
			    
			    /**** add it to corpus_hash if necessary or increment the counter ****/
			    
			    Integer count;
			    if( (count = (Integer)corpus_hash.get(word)) == null ){
				
				
				/**** add it to corpus_hash ****/
				
				corpus_hash.put(word,new Integer(1));
			    }
			    else{
				
				/**** increment counter in doc_freq ****/
				
				Integer newCount = new Integer(count.intValue()+1);
				corpus_hash.put(word,newCount);
			    }
			}
		    }
		}
		dis.close();
		
	    } catch(IOException ex){ex.printStackTrace();}
	}
    }
    
    private Vector getDocList(String input_dir_name) {

	File input_dir = new File(input_dir_name);
	if (!input_dir.exists()) {
	    System.err.println("CKIndexer Error: input directory "+input_dir_name+ " does not exist!");
	    return null;
	}

	Vector file_vector = new Vector();
	processDirectory(input_dir, file_vector);
	return file_vector;
    }
    private void processDirectory(File dir, Vector file_list) {
	File [] files = dir.listFiles();
	for (int i=0; i<files.length; i++) {
	    if (files[i].isDirectory()) {
		processDirectory(files[i], file_list);
	    } else {
		file_list.add(files[i].getPath());
	    }
	}

    }
//     private Vector readDocList(String list)
//     {
// 	list = inputDir + "/" + "docs.lst";
	
// 	File docList = new File(list);
	
// 	if (!docList.exists()){
// 	    System.out.println("No documents found");
// 	    System.exit(0);
// 	}
	
// 	Vector docs = new Vector();
	
// 	try{
//             BufferedReader dis = new BufferedReader(new InputStreamReader(new FileInputStream(docList)));
// 	    String line = "";
//             while( (line=dis.readLine())!=null ){
// 		docs.addElement(new String(inputDir+"/" + line));
// 	    }
//             dis.close();
// 	} catch(IOException ex){ex.printStackTrace();}
	
// 	return docs;
//     }
    
    
    private void printToFile(Object obj, String name){
	
	FileOutputStream fos = null;
	ObjectOutputStream out = null;
	
	try{
	    fos = new FileOutputStream(name);
	    out = new ObjectOutputStream(fos);
	    out.writeObject(obj);
	    out.close();
	}
	catch(IOException ex){
	    ex.printStackTrace();
	}
    }
}


	/*Hgram hg = new Hgram();
	hg.setBinNumber(50);
	hg.setLabel("Document frequency");
	hg.setData(freqs_f,Hgram.RAW_DATA);
	hg.update();*/

 /*    private Map.Entry[] sort(Hashtable h)
    {
	Set set = h.entrySet();
	Map.Entry [] entries = 
	    (Map.Entry[])set.toArray(
				     new Map.Entry[set.size()]);
	Arrays.sort(entries, new Comparator() {
		public int compare(Object o1, Object o2) {
		    int i1 = ((Integer)((Map.Entry)o1).getValue()).intValue();
		    int i2 = ((Integer)((Map.Entry)o2).getValue()).intValue();
		    if( i1 < i2 ) return 0;
		    else return 1;
		}
		public boolean equals(Object that) {
		    return this==that;
		}
	    });
	return entries;
	}*/
